# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-04-07 16:17
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('journal', '0007_auto_20170407_1325'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='userinfo',
            name='deleted',
        ),
    ]
